from enum import Enum, auto
from tkinter import StringVar, Button, Entry, Frame
from typing import Union, Dict

import diffeq_solver_tk.messages.common_messages as common_messages
import diffeq_solver_tk.messages.heat_equation_messages as messages
from diffeq_solver_tk.differential_equation_metadata import BoundaryConditions, BoundaryCondition, HeatEquationMetadata, BoundaryType
from diffeq_solver_tk.forms.differential_equation_form import DifferentialEquationForm
from diffeq_solver_tk.forms.equation_form_builder import EquationFormBuilder
from diffeq_solver_tk.services.heat_equation_service import HeatEquationService


class HeatEquationFields(Enum):
    LEFT_BOUNDARY_TYPE = auto()
    RIGHT_BOUNDARY_TYPE = auto()
    LEFT_BOUNDARY_VALUES = auto()
    RIGHT_BOUNDARY_VALUES = auto()
    INITIAL_VALUES = auto()
    SOURCE = auto()
    ALPHA = auto()
    LENGTH = auto()
    TIME = auto()
    SAMPLES = auto()


class HeatEquationForm(DifferentialEquationForm):
    equation_service: HeatEquationService
    field_entry_map: Dict[HeatEquationFields, Union[Entry, StringVar]]
    solve_button: Button
    export_button: Button
    render_plot_button: Button
    toggle_animation_button: Button

    def __init__(self, frame: Frame, canvas, equation_service):
        DifferentialEquationForm.__init__(self, frame, canvas, equation_service)

    def build_form(self):
        builder = EquationFormBuilder[HeatEquationFields](self)
        self.build_entries(builder)
        self.build_buttons(builder)

    def build_entries(self, builder: EquationFormBuilder[HeatEquationFields]):
        builder.build_boundary_type_section(HeatEquationFields.LEFT_BOUNDARY_TYPE,
                                            HeatEquationFields.RIGHT_BOUNDARY_TYPE)
        builder.build_entry_row(HeatEquationFields.LEFT_BOUNDARY_VALUES,
                                messages.left_boundary_values,
                                messages.left_boundary_type_symbol, 2)
        builder.build_entry_row(HeatEquationFields.RIGHT_BOUNDARY_VALUES,
                                messages.right_boundary_values,
                                messages.right_boundary_type_symbol, 3)
        builder.build_entry_row(HeatEquationFields.INITIAL_VALUES,
                                messages.initial_values,
                                messages.initial_values_symbol, 4)
        builder.build_entry_row(HeatEquationFields.SOURCE,
                                messages.source_term,
                                messages.source_term_symbol, 5)
        builder.build_entry_row(HeatEquationFields.ALPHA,
                                messages.diffusivity,
                                messages.diffusivity_symbol, 6)
        builder.build_entry_row(HeatEquationFields.LENGTH,
                                messages.length,
                                messages.length_symbol, 7)
        builder.build_entry_row(HeatEquationFields.TIME,
                                messages.time_interval,
                                messages.time_interval_symbol, 8)
        builder.build_entry_row(HeatEquationFields.SAMPLES,
                                messages.samples,
                                messages.samples_symbol, 9)
        self.field_entry_map = builder.get_field_entry_map()

    def build_buttons(self, builder: EquationFormBuilder[HeatEquationFields]):
        self.solve_button = builder.create_button(common_messages.solve, callback=self.solve_equation)
        self.export_button = builder.create_button(common_messages.export, callback=self.export_solution)
        self.render_plot_button = builder.create_button(common_messages.show_plot, callback=self.handle_render_plot)
        self.toggle_animation_button = builder.create_button(common_messages.play,
                                                             callback=self.handle_toggle_animation)
        self.solve_button.grid(row=10, column=2, pady=6, sticky="w")

    def get_equation_metadata(self):
        left_boundary_type = BoundaryType(self.field_entry_map[HeatEquationFields.LEFT_BOUNDARY_TYPE].get())
        right_boundary_type = BoundaryType(self.field_entry_map[HeatEquationFields.RIGHT_BOUNDARY_TYPE].get())
        boundary_conditions = BoundaryConditions(
            BoundaryCondition(left_boundary_type,
                              self.field_entry_map[HeatEquationFields.LEFT_BOUNDARY_VALUES].get()),
            BoundaryCondition(right_boundary_type,
                              self.field_entry_map[HeatEquationFields.RIGHT_BOUNDARY_VALUES].get()))
        source = self.field_entry_map[HeatEquationFields.SOURCE].get()
        return HeatEquationMetadata(
            boundary_conditions,
            float(self.field_entry_map[HeatEquationFields.LENGTH].get()),
            int(self.field_entry_map[HeatEquationFields.SAMPLES].get()),
            float(self.field_entry_map[HeatEquationFields.TIME].get()),
            float(self.field_entry_map[HeatEquationFields.ALPHA].get()),
            self.field_entry_map[HeatEquationFields.INITIAL_VALUES].get(),
            source if source else "0"
        )

    def on_solve(self):
        self.render_plot_button.grid(row=10, column=3, pady=6, sticky="e")
        self.toggle_animation_button.grid(row=11, column=3, pady=6, sticky="e")
        self.export_button.grid(row=12, column=3, pady=6, sticky="e")

    def handle_render_plot(self):
        self.toggle_animation_button.configure(text=common_messages.play)
        self.equation_service.render_current_solution()
        self.canvas.draw()

    def handle_toggle_animation(self):
        self.equation_service.toggle_animation()
        self.canvas.draw()
        if self.equation_service.is_animation_playing():
            self.toggle_animation_button.configure(text=common_messages.pause)
        else:
            self.toggle_animation_button.configure(text=common_messages.play)

    def reset(self):
        self.equation_service.clear_solution()
        self.equation_service.clear_animation()
        self.canvas.draw()
        self.render_plot_button.grid_forget()
        self.toggle_animation_button.grid_forget()
        self.export_button.grid_forget()



import timeit
from typing import Callable

import numpy as np
from matplotlib.figure import Figure

from diffeq_solver_tk.differential_equation_metadata import OrdinaryDifferentialEquationMetadata
from diffeq_solver_tk.services.differential_equation_service import OrdinaryDifferentialEquationService


class SecondOrderODEService(OrdinaryDifferentialEquationService):
    def __init__(self, main_figure: Figure):
        super().__init__(main_figure)

    def compute_solution(self, metadata: OrdinaryDifferentialEquationMetadata) -> np.ndarray:
        start_time = timeit.default_timer()
        N = metadata.samples
        dt = metadata.time / (N - 1)
        initial_value, derivative = metadata.initial_derivatives
        source: Callable[[float, float, float], float] = lambda t, x, y: eval(metadata.source)
        solution = np.zeros(N)
        solution[0] = initial_value
        for i in range(1, N):
            derivative_delta = source((i - 1) * dt, solution[i - 1], derivative) * dt
            derivative = derivative + derivative_delta
            solution[i] = solution[i - 1] + derivative * dt
        total_time = timeit.default_timer() - start_time
        print(f"computing second order ode solution took {total_time} seconds")
        return solution

    def validate_metadata(self, metadata: OrdinaryDifferentialEquationMetadata):
        if metadata.samples <= 1 or metadata.time == 0:
            raise Exception("invalid second order ode metadata")



from enum import Enum, auto
from tkinter import Entry, Frame, Button
from typing import Dict

import diffeq_solver_tk.messages.common_messages as common_messages
import diffeq_solver_tk.messages.first_order_ode_messages as messages
from diffeq_solver_tk.differential_equation_metadata import OrdinaryDifferentialEquationMetadata
from diffeq_solver_tk.forms.differential_equation_form import DifferentialEquationForm
from diffeq_solver_tk.forms.equation_form_builder import EquationFormBuilder
from diffeq_solver_tk.services.first_order_ode_service import FirstOrderODEService


class FirstOrderODEFields(Enum):
    SOURCE = auto()
    INITIAL_VALUE = auto()
    TIME = auto()
    SAMPLES = auto()


class FirstOrderODEForm(DifferentialEquationForm):
    equation_service: FirstOrderODEService
    field_entry_map: Dict[FirstOrderODEFields, Entry]
    export_button: Button

    def __init__(self, frame: Frame, canvas, equation_service: FirstOrderODEService):
        DifferentialEquationForm.__init__(self, frame, canvas, equation_service)

    def build_form(self):
        builder: EquationFormBuilder[FirstOrderODEFields] = EquationFormBuilder[FirstOrderODEFields](self)
        builder.build_entry_row(FirstOrderODEFields.SOURCE,
                                messages.source_term,
                                messages.source_term_symbol, 0)
        builder.build_entry_row(FirstOrderODEFields.INITIAL_VALUE,
                                messages.initial_value,
                                messages.initial_value_symbol, 1)
        builder.build_entry_row(FirstOrderODEFields.TIME,
                                messages.time_interval,
                                messages.time_interval_symbol, 2)
        builder.build_entry_row(FirstOrderODEFields.SAMPLES,
                                messages.samples,
                                messages.samples_symbol, 3)
        self.field_entry_map = builder.get_field_entry_map()
        builder.create_button(common_messages.solve,
                              callback=self.solve_equation).grid(row=5, column=2, pady=10, sticky="w")
        self.export_button = builder.create_button(common_messages.export, callback=self.export_solution)

    def get_equation_metadata(self):
        return OrdinaryDifferentialEquationMetadata(
            self.field_entry_map.get(FirstOrderODEFields.SOURCE).get(),
            [float(self.field_entry_map.get(FirstOrderODEFields.INITIAL_VALUE).get())],
            int(self.field_entry_map.get(FirstOrderODEFields.SAMPLES).get()),
            float(self.field_entry_map.get(FirstOrderODEFields.TIME).get())
        )

    def on_solve(self):
        self.export_button.grid(row=6, column=2, sticky="w")

    def reset(self):
        self.equation_service.clear_solution()
        self.export_button.grid_forget()
        self.canvas.draw()



source_term = "Source term"
initial_value = "Initial value"
initial_derivative = "Initial Derivative"
time_interval = "Time interval"
samples = "Samples"

source_term_symbol = "f(t, x, y=x')"
initial_value_symbol = "x(0)"
initial_derivative_symbol = "x'(0)"
time_interval_symbol = "T"
samples_symbol = "N"



from tkinter import Tk

import diffeq_solver_tk.messages.common_messages as common_messages
from .main_view import MainView


def main():
    app = Tk(className=common_messages.app_name)
    app.resizable(width=False, height=False)
    # app.iconbitmap('assets/icon.ico')
    app.geometry("1100x600")
    MainView(app).pack(side="top", fill="both", expand=True)
    app.mainloop()


if __name__ == "__main__":
    main()



left_boundary_values = "Left boundary values"
right_boundary_values = "Right boundary values"
initial_values = "Initial Values"
source_term = "Source term (optional)"
diffusivity = "Diffusivity"
length = "Length"
time_interval = "Time interval"
samples = "Samples"

left_boundary_type_symbol = "φ₁(t)"
right_boundary_type_symbol = "φ₂(t)"
initial_values_symbol = "u(0, x)"
source_term_symbol = "s(t, x)"
heat_equation_initial_value_symbol = "u(0, x)"
heat_equation_source_term_symbol = "s(t, x)"
diffusivity_symbol = "α"
length_symbol = "L"
time_interval_symbol = "T"
samples_symbol = "N"
